import logging
import datetime

from pystibmivb.client import AbstractSTIBAPIClient
from pystibmivb.domain.passages import Passage
from pystibmivb.service.ShapefileService import ShapefileService

LOGGER = logging.getLogger(__name__)

LANGS = {"fr", "nl"}


class InvalidLineFilterException(Exception):
    pass


class STIBService:
    def __init__(self, stib_api_client: AbstractSTIBAPIClient):
        self._shapefile_service = ShapefileService(stib_api_client)
        self.api_client = stib_api_client

    async def get_passages(self, stop_name: str, line_filters: list = None, max_passages: int = None, lang_stop_name: str = None, lang_message: str = None,
                           now: datetime.datetime = datetime.datetime.now()):
        stop_infos = await self._shapefile_service.get_stop_infos(stop_name)

        if lang_message is None or lang_message not in LANGS:
            lang_message = "fr"
        if lang_stop_name is None or lang_stop_name not in LANGS:
            lang_stop_name = "fr"


        atomic_stop_infos = stop_infos.get_atomic_stop_infos(line_filters)
        if len(atomic_stop_infos) < 1:
            LOGGER.error("Invalid line filter for stop " + str(stop_name) + " Known infos:" + str(
                stop_infos) + " Provided line filter:" + str(line_filters))
            raise InvalidLineFilterException()
        passages = []
        is_only_end_of_service = True
        raw_passages = await self.api_client.api_call_passingTimeByPoint_for_stop_ids(stop_infos.get_stop_ids(line_filters))
        for point in raw_passages["points"]:
            for json_passage in point["passingTimes"]:
                message = ""
                try:
                    message = json_passage["message"][lang_message]
                except KeyError:
                    pass
                try:
                    if message.upper() == "FIN DE SERVICE" or message.upper() == "EINDE DIENST":
                        next_passage = await self._shapefile_service.get_next_stop_passage(point["pointId"], now)
                        passages.append(Passage(stop_id=point["pointId"],
                                                lineId=json_passage["lineId"],
                                                destination="",
                                                expectedArrivalTime=next_passage.strftime("%Y-%m-%dT%H:%M:%S"),
                                                lineInfos=await self._shapefile_service.get_line_info(
                                                    json_passage["lineId"]),
                                                message=message,
                                                now=now))
                    else:
                        is_only_end_of_service = False
                        passages.append(Passage(stop_id=point["pointId"],
                                                lineId=json_passage["lineId"],
                                                destination=json_passage["destination"][lang_stop_name],
                                                expectedArrivalTime=json_passage.get("expectedArrivalTime", None),
                                                lineInfos=await self._shapefile_service.get_line_info(
                                                    json_passage["lineId"]),
                                                message=message,
                                                now=now))
                except KeyError as ke:
                    LOGGER.error(
                        "Error while parsing response from STIB. Raw response is: " + str(raw_passages))
                    raise ke
        if is_only_end_of_service or len(passages) == 0:
            raised_passages = []
            message = "Information from Schedule"
            if lang_message == 'fr':
                message = "Information venant du planning"
            elif lang_message == 'nl':
                message = "Informatie uit de planning"
            for atomic in atomic_stop_infos:
                next_passage = await self._shapefile_service.get_next_stop_passage(atomic.get_stop_id(), now)
                raised_passages.append(Passage(stop_id=atomic.get_stop_id(),
                                               lineId=atomic.get_line_nr(),
                                               destination=atomic.get_destination(),
                                               expectedArrivalTime=next_passage.strftime("%Y-%m-%dT%H:%M:%S"),
                                               lineInfos=await self._shapefile_service.get_line_info(
                                                   atomic.get_line_nr()),
                                               message=message,
                                               now=now))
            raise NoScheduleFromAPIException(sorted(raised_passages))

        passages = sorted(passages)
        if max_passages is not None:
            passages = passages[0:max_passages]
        return passages


class NoScheduleFromAPIException(Exception):
    def __init__(self, passages):
        """ This exception provides information about the next scheduled passage.
        The fact that an exception is raised allows to differentiate passages returned from API or assumed values from GTFS
        It could be used to specify the user that this info is assumed instead of "official from API" """
        self.passages = passages

    def get_next_passages(self):
        return self.passages



class LineInfo:
    def __init__(self, line_nr, line_type, line_color, line_text_color):
        self.line_nr = line_nr
        self.line_type = line_type
        self.line_color = line_color
        self.line_text_color = line_text_color

    def get_line_nr(self):
        return self.line_nr

    def get_line_type(self):
        return self.line_type

    def get_line_color(self):
        return self.line_color

    def get_line_text_color(self):
        return self.line_text_color

    def __str__(self):
        return f"{self.line_nr}: {self.line_type}: {self.line_color}: {self.line_text_color}"



import logging
import os
import tempfile
import zipfile

import shapefile
import datetime

from pystibmivb.client import AbstractSTIBAPIClient
from pystibmivb.domain.line import LineInfo
from pystibmivb.domain.stop import StopInfo

LOGGER = logging.getLogger(__name__)

ENDPOINT_SHAPEFILES = '/Files/2.0/Shapefiles'
ENDPOINT_GTFS = '/Files/2.0/Gtfs'

SEP = os.sep
SHAPEFILESFOLDERPATH = tempfile.gettempdir() + SEP + "stibmivbshapefiles"
SHAPEFILES_ZIP_FILENAME = "shapefiles.zip"
SHAPEFILES_ZIP_FILEPATH = SHAPEFILESFOLDERPATH + SEP + SHAPEFILES_ZIP_FILENAME
GTFS_ZIP_FILENAME = "gtfs.zip"
GTFS_ZIP_FILEPATH = SHAPEFILESFOLDERPATH + SEP + GTFS_ZIP_FILENAME
LINES_FILENAME = "ACTU_LINES"
LINES_FILEPATH = SHAPEFILESFOLDERPATH + SEP + LINES_FILENAME
LINES_GTFS_FILENAME = "routes.txt"
LINES_GTFS_FILEPATH = SHAPEFILESFOLDERPATH + SEP + LINES_GTFS_FILENAME
STOPS_FILENAME = "ACTU_STOPS"
STOPS_FILEPATH = SHAPEFILESFOLDERPATH + SEP + STOPS_FILENAME
STOPS_GTFS_FILENAME = "stops.txt"
STOPS_GTFS_FILEPATH = SHAPEFILESFOLDERPATH + SEP + STOPS_GTFS_FILENAME
STOPS_SCHEDULES_GTFS_FILENAME = "stop_times.txt"
STOPS_SCHEDULES_GTFS_FILEPATH = SHAPEFILESFOLDERPATH + SEP + STOPS_SCHEDULES_GTFS_FILENAME
TIMESTAMPFILENAME = "timestamp"
TIMESTAMPFILEPATH = SHAPEFILESFOLDERPATH + SEP + TIMESTAMPFILENAME
DELTA_MAX_TIMESTAMP = 1 * 60 * 60 * 24 * 7  # 1 week


def _recur_find_next_passage(date: datetime.datetime, target_time: str, array: list, start: int,
                             end: int) -> datetime.datetime:
    mid = (start + end) // 2
    if mid == start:
        if target_time < array[end]:
            h, m, s = array[end].split(':')
            return datetime.datetime(year=date.year, month=date.month, day=date.day + int(h)//24, hour=int(h)%24,
                                     minute=int(m), second=int(s))
        else:
            h, m, s = array[0].split(':')
            return datetime.datetime(year=date.year, month=date.month, day=date.day, hour=int(h), minute=int(m),
                                     second=int(s)) + datetime.timedelta(days=1)
    if target_time < array[mid]:
        return _recur_find_next_passage(date, target_time, array, start, mid)
    else:  # array[mid] <= target_time
        return _recur_find_next_passage(date, target_time, array, mid, end)


class ShapefileService:
    def __init__(self, stib_api_client: AbstractSTIBAPIClient):
        self.api_client = stib_api_client
        self.lines_cache = {}
        self.stops_cache = {}
        self.stops_schedules_cache = {}

    async def get_line_text_color(self, line_nr):
        await self._refresh_files()
        with open(LINES_GTFS_FILEPATH, 'r') as lines_file:
            for info_line in lines_file:
                splitted_info_line = info_line.split(',')
                if splitted_info_line[1].strip() == str(line_nr):
                    return "#" + splitted_info_line[-1].strip()
        return "#E1337E"

    async def get_line_info(self, line_nr: int) -> LineInfo:
        await self._refresh_files()
        if line_nr not in self.lines_cache.keys():
            sf = shapefile.Reader(LINES_FILEPATH)
            for record in sf.records():
                record = record.as_dict()
                current_line_nr, line_type = record["LIGNE"][:-1], record["LIGNE"][-1:]
                if int(line_nr) == int(current_line_nr):
                    line_color = record["COLOR_HEX"]
                    line_text_color = await self.get_line_text_color(line_nr)
                    self.lines_cache[line_nr] = LineInfo(int(current_line_nr), line_type.upper(), line_color,
                                                         line_text_color)
                    break
        return self.lines_cache[line_nr]

    async def get_stop_schedule(self, stop_id: str):
        await self._refresh_files()
        if stop_id not in self.stops_schedules_cache.keys():
            LOGGER.info("Stop schedule for stop_id:" + stop_id + " not cached, fetching it in GTFS files...")
            res = []
            with open(STOPS_SCHEDULES_GTFS_FILEPATH, 'r') as gtfs_schedules:
                for schedule_part in gtfs_schedules:
                    schedule_part_splitted = schedule_part.strip().split(',')
                    if schedule_part_splitted[3].upper() == stop_id.upper():
                        res.append(schedule_part_splitted[2].strip())
            self.stops_schedules_cache[stop_id] = sorted(res)
            LOGGER.info("Stop schedule for stop_id:" + stop_id + " successfully cached!")
        return self.stops_schedules_cache[stop_id]

    async def get_next_stop_passage(self, stop_id: str, time: datetime.datetime) -> datetime.datetime:
        stop_schedule = await self.get_stop_schedule(stop_id)
        return _recur_find_next_passage(time, time.strftime("%H:%M:%S"), stop_schedule, 0, len(stop_schedule) - 1)

    async def get_stop_infos(self, stop_name: str) -> StopInfo:
        await self._refresh_files()
        if stop_name not in self.stops_cache.keys():
            sf = shapefile.Reader(STOPS_FILEPATH)
            res = StopInfo(stop_name)
            for record in sf.records():
                record = record.as_dict()
                if record["alpha_fr"].upper() == stop_name.upper() or record["alpha_nl"].upper() == stop_name.upper()\
                        or record["descr_fr"].upper() == stop_name.upper() or record["descr_nl"].upper() == stop_name.upper():
                    res.add_stop(record["stop_id"], record["numero_lig"], record["variante"], record["terminus"])
                    res.add_line_info(await self.get_line_info(record["numero_lig"]))
            res.set_locations(await self.get_stop_locations(res.get_stop_ids()))
            self.stops_cache[stop_name] = res
        return self.stops_cache[stop_name]

    async def _refresh_files(self):
        if self._must_update_files():
            import time
            LOGGER.info("Shapefiles validity outdated, updating them...")
            with open(TIMESTAMPFILEPATH, 'w') as f:
                f.write(str(time.time()))
            await self._refresh_shapefiles()
            await self._refresh_gtfs()

    async def _refresh_gtfs(self):
        zipped_data = await self.api_client.api_call(ENDPOINT_GTFS)
        if zipped_data is not None:
            # save data to disk
            zip_path = GTFS_ZIP_FILEPATH
            LOGGER.info("Saving new zip of gtfs to " + str(zip_path))
            with open(zip_path, 'wb') as output:
                output.write(zipped_data)
                output.close()

            # extract the data
            zfobj = zipfile.ZipFile(zip_path)
            for name in zfobj.namelist():
                uncompressed = zfobj.read(name)
                name = name.split('/')[-1]

                # save uncompressed data to disk
                output_filename = SHAPEFILESFOLDERPATH + SEP + name
                LOGGER.info("Saving extracted file to " + str(output_filename))
                with open(output_filename, 'wb') as output:
                    output.write(uncompressed)

            # TODO os.remove(zip_filename)
            LOGGER.info("Finished updating GTFS!")
            self.lines_cache = {}
            self.stops_schedules_cache = {}
        else:
            LOGGER.error("Unable to update GTFS...")

    async def _refresh_shapefiles(self):
        """ Get most recent file info if not in local cache (api for files can be called only once per minute.
        These file change only 2 or 3 times per year thus we will invalidate them after one week.
        To force update simply delete the timestamp file."""
        zipped_data = await self.api_client.api_call(ENDPOINT_SHAPEFILES)
        if zipped_data is not None:
            # save data to disk
            zip_path = SHAPEFILES_ZIP_FILEPATH
            LOGGER.info("Saving new zip of shapefiles to " + str(zip_path))
            with open(zip_path, 'wb') as output:
                output.write(zipped_data)
                output.close()

            # extract the data
            zfobj = zipfile.ZipFile(zip_path)
            for name in zfobj.namelist():
                uncompressed = zfobj.read(name)
                name = name.split('/')[-1]

                # save uncompressed data to disk
                output_filename = SHAPEFILESFOLDERPATH + SEP + name
                LOGGER.info("Saving extracted file to " + str(output_filename))
                with open(output_filename, 'wb') as output:
                    output.write(uncompressed)

            # TODO os.remove(zip_filename)
            LOGGER.info("Finished updating Shapefiles!")
            self.stops_cache = {}
        else:
            LOGGER.error("Unable to update Shapefiles...")

    def _must_update_files(self):
        import time
        must_update = False
        if not os.path.isdir(SHAPEFILESFOLDERPATH):
            LOGGER.info("Shapefile folder not existing, creating it...")
            must_update = True
            os.mkdir(SHAPEFILESFOLDERPATH)
        if not os.path.isfile(TIMESTAMPFILEPATH):
            LOGGER.info("Shapefile timestamp file not existing, creating it...")
            with open(TIMESTAMPFILEPATH, 'w') as f:
                f.write(str(time.time()))
            must_update = True
        with open(TIMESTAMPFILEPATH, 'r') as f:
            timestamp = int(f.read().split(".")[0])
            now = time.time()
            if now - timestamp > DELTA_MAX_TIMESTAMP:
                must_update = True
                LOGGER.info(
                    f"Delta since last update is {now - timestamp} which is greater than {DELTA_MAX_TIMESTAMP}. Invalidating files...")
        return must_update

    async def get_stop_locations(self, stop_ids: list):
        await self._refresh_files()
        res = {}
        for stop_id in stop_ids:
            res[stop_id] = await self.get_stop_location(stop_id)
        return res

    async def get_stop_location(self, stop_id: str):
        await self._refresh_files()
        with open(STOPS_GTFS_FILEPATH, 'r') as stops_file:
            for info_stop in stops_file:
                splitted_info_line = info_stop.split(',')
                if splitted_info_line[0].strip() == stop_id:
                    return {"lat": float(splitted_info_line[4]), "lon": float(splitted_info_line[5])}
        return {"lat": 50.847180, "lon": 4.361610}



class AtomicStop:
    def __init__(self, stop_id, line_nr, variante, terminus):
        self.stop_id = stop_id
        self.line_nr = line_nr
        self.variante = variante
        self.terminus = terminus

    def get_stop_id(self):
        return self.stop_id

    def get_line_nr(self):
        return self.line_nr

    def get_destination(self):
        return self.terminus

    def get_variante(self):
        return self.variante

    def __str__(self):
        return f"{self.stop_id} {self.line_nr} {self.variante} {self.terminus}"

    def __repr__(self):
        return super().__repr__() + self.__str__()


class StopInfo:
    def __init__(self, stop_name):
        self.stop_name = stop_name
        self.lines = []
        self.line_infos = {}
        self.locations = {}

    def add_stop(self, stop_id, line_nr, variante, terminus):
        self.lines.append(AtomicStop(stop_id, line_nr, variante, terminus))

    def add_line_info(self, line_info):
        self.line_infos[line_info.get_line_nr()] = line_info

    def get_line_info(self, line_nr):
        return self.line_infos[line_nr]

    def get_atomic_stop_infos(self, line_filters=None) -> list:
        res = []
        if line_filters is not None and len(line_filters) > 0:
            line_filter_dict = {}
            for line_nr, line_variant_or_dest in line_filters:
                line_filter_dict[line_nr] = line_filter_dict.get(line_nr, [])
                line_filter_dict[line_nr].append(
                    line_variant_or_dest.upper() if isinstance(line_variant_or_dest, str) else line_variant_or_dest)
            for atomic_stop_info in self.lines:
                if atomic_stop_info.get_line_nr() in line_filter_dict.keys():  # the line nr is included in line_filter
                    # now we check for direction
                    if atomic_stop_info.get_destination().upper() in line_filter_dict[atomic_stop_info.get_line_nr()]:
                        res.append(atomic_stop_info)
                    elif atomic_stop_info.get_variante() in line_filter_dict[atomic_stop_info.get_line_nr()]:
                        res.append(atomic_stop_info)
        else:
            res = self.lines
        return res

    def get_stop_ids(self, line_filters=None) -> list:
        res = []
        if line_filters is not None and len(line_filters) > 0:
            line_filter_dict = {}
            for line_nr, line_variant_or_dest in line_filters:
                line_filter_dict[line_nr] = line_filter_dict.get(line_nr, [])
                line_filter_dict[line_nr].append(
                    line_variant_or_dest.upper() if isinstance(line_variant_or_dest, str) else line_variant_or_dest)
            for atomic_stop_info in self.lines:
                if atomic_stop_info.get_line_nr() in line_filter_dict.keys():  # the line nr is included in line_filter
                    # now we check for direction
                    if atomic_stop_info.get_destination().upper() in line_filter_dict[atomic_stop_info.get_line_nr()]:
                        res.append(atomic_stop_info)
                    elif atomic_stop_info.get_variante() in line_filter_dict[atomic_stop_info.get_line_nr()]:
                        res.append(atomic_stop_info)
        else:
            res = self.lines
        return [ln.get_stop_id() for ln in res]

    def __str__(self):
        return f"{self.stop_name}({self.get_location()}): {[str(k) for k in self.lines]}: {[str(v) for v in self.line_infos.values()]}"

    def set_locations(self, locs):
        self.locations = locs

    def get_locations(self):
        """ https://xkcd.com/2170/ """
        return self.locations

    def get_location(self):
        """ https://xkcd.com/2170/ """
        latitudes = [loc["lat"] for loc in self.locations.values()]
        longitudes = [loc["lon"] for loc in self.locations.values()]
        return {"lat": sum(latitudes)/len(latitudes), "lon": sum(longitudes)/len(longitudes)}

    def get_lines(self, lines_filter=None):
        return set([asi.get_line_nr() for asi in self.get_atomic_stop_infos(lines_filter)])

    def get_lines_with_destinations(self, lines_filter=None):
        return [(asi.get_line_nr(), asi.get_destination()) for asi in self.get_atomic_stop_infos(lines_filter)]



import math
from datetime import *


def get_time_delta(t1, t2):
    if t1 is None or t2 is None:
        return None
    if not type(t1) == type(datetime.now()):
        t1 = datetime.strptime(str(t1), "%Y-%m-%dT%H:%M:%S")
    t2 = datetime.strptime(t2.split("+")[0], "%Y-%m-%dT%H:%M:%S")  # 2020-02-08T14:40:00+01:00
    delta = t2.timestamp() - t1.timestamp()
    ret = max(0, int(math.floor(delta)))
    return ret



from datetime import *

from .line import LineInfo
from .util import get_time_delta


class Passage(dict):
    def __init__(self, stop_id: int, lineId: int, destination: str, expectedArrivalTime: str, lineInfos: LineInfo,
                 message: str = "", now: datetime = datetime.now()):
        dict.__init__(self)
        self["stop_id"] = stop_id
        self["line_id"] = lineId
        self["destination"] = destination
        self["line_color"] = lineInfos.get_line_color()
        self["line_text_color"] = lineInfos.get_line_text_color()
        self["line_type"] = lineInfos.get_line_type()
        self["message"] = message
        if expectedArrivalTime is not None:
            self["expected_arrival_time"] = expectedArrivalTime
            delta = get_time_delta(now, expectedArrivalTime)
            self["arriving_in"] = {"min": delta // 60, "sec": delta % 60}

    def __lt__(self, other):
        if type(other) != type(self):
            return super(self).__lt__(other)
        other_arriving_in = other.get("arriving_in", None)
        if other_arriving_in is not None:
            self_arriving_in = self.get("arriving_in", None)
            if self_arriving_in is not None:
                if self_arriving_in["min"] < other_arriving_in["min"]:
                    return True
                elif other_arriving_in["min"] < self_arriving_in["min"]:
                    return False
                else:  # == we compare seconds
                    if self_arriving_in["sec"] < other_arriving_in["sec"]:
                        return True
                    else:
                        return False
            else:
                return False
        else:
            return True

    def __eq__(self, other):
        if type(other) != type(self):
            return super(self).__eq__(other)
        other_arriving_in = other.get("arriving_in", None)
        if other_arriving_in is not None:
            self_arriving_in = self.get("arriving_in", None)
            if self_arriving_in is not None:
                return self_arriving_in["min"] == other_arriving_in["min"] \
                       and self_arriving_in["sec"] == other_arriving_in["sec"]
            else:
                return False
        else:
            return False



"""Common attributes and functions."""
import asyncio
import base64
import logging
import time
from abc import ABC, abstractmethod

import aiohttp
import async_timeout
from yarl import URL

LOGGER = logging.getLogger(__name__)

API_BASE_URL = 'https://opendata-api.stib-mivb.be'
PASSING_TIME_BY_POINT_SUFFIX = "/OperationMonitoring/4.0/PassingTimeByPoint/"


class OAuthTokenManager(object):
    def __init__(
            self,
            loop,
            session,
            client_id,
            client_secret,
            url,
            *,
            renew_pad_secs=60,
            **parameters
    ):
        self.session = session
        self.client_secret = client_secret
        self.client_id = client_id
        self.loop = loop
        self.url = url
        self.parameters = parameters
        self.renew_pad_secs = renew_pad_secs

        self._token = None
        self._exp = None

    async def login(self):
        headers = {}
        headers[
            'Authorization'] = f'Basic {str(base64.b64encode((self.client_id + ":" + self.client_secret).encode("utf-8")), "utf-8")}'
        try:
            async with async_timeout.timeout(5, loop=self.loop):
                response = await self.session.post(data={"grant_type": "client_credentials"},
                                                   url=self.url,
                                                   headers=headers)
                body = {}
                if response.content_type == "application/json":
                    body = await response.json()
                else:
                    raise aiohttp.ClientError("Unexpected content type: got " + response.content_type)
                self._token = body['access_token']
                self._exp = time.time() + body['expires_in'] - self.renew_pad_secs
        except aiohttp.ClientError as error:
            LOGGER.error("Error fetching token for STIB/MIVB : %s", error)
        except asyncio.TimeoutError as error:
            LOGGER.debug("Timeout connecting to STIB/MIVB API: %s", error)

    def is_token_valid(self):
        return self._token and self._exp and time.time() < self._exp

    @property
    def token(self):
        return self._token


class AbstractSTIBAPIClient(ABC):
    @abstractmethod
    async def api_call(self, endpoint_suffix: str, additional_headers=None):
        pass

    @abstractmethod
    async def api_call_passingTimeByPoint_for_stop_id(self, point_id: str) -> dict:
        pass

    @abstractmethod
    async def api_call_passingTimeByPoint_for_stop_ids(self, point_ids: list) -> dict:
        pass


class STIBAPIClient(AbstractSTIBAPIClient):
    """A class for common functions."""

    def __init__(self,
                 loop: asyncio.events.AbstractEventLoop,
                 session: aiohttp.ClientSession,
                 client_id: str,
                 client_secret: str):
        """Initialize the class."""
        self.loop = loop
        self.session = session
        if self.session is None:
            self.session = aiohttp.ClientSession()
        self.client_id = client_id
        self.client_secret = client_secret
        self.token_manager = OAuthTokenManager(self.loop, self.session,
                                               self.client_id, self.client_secret,
                                               API_BASE_URL + '/token', )

    async def async_get_access_token(self) -> str:
        """Return a valid access token."""
        if self.token_manager.is_token_valid():
            return self.token_manager.token

        await self.token_manager.login()

        return self.token_manager.token

    async def api_call_passingTimeByPoint_for_stop_id(self, point_id: str) -> dict:
        return await self.api_call(PASSING_TIME_BY_POINT_SUFFIX+point_id)

    async def _api_call_passingTimeByPoint_for_10_stop_ids(self, point_ids: list) -> dict:
        return await self.api_call(PASSING_TIME_BY_POINT_SUFFIX + "%2C".join(point_ids))

    async def api_call_passingTimeByPoint_for_stop_ids(self, point_ids: set) -> dict:
        res = {"points": []}
        subset_point_ids = []
        for i in range(1, len(point_ids)+1):
            subset_point_ids.append(point_ids.pop())
            if i % 10 == 0 or len(point_ids) == 0:
                subset_res = await self._api_call_passingTimeByPoint_for_10_stop_ids(subset_point_ids)
                res["points"].extend(subset_res["points"])
                subset_point_ids = []
        return res

    async def api_call(self, endpoint_suffix: str, additional_headers=None):
        if additional_headers is None:
            additional_headers = {}

        """Call the API."""
        headers = {'Authorization': "Bearer " + await self.async_get_access_token()}
        headers.update(additional_headers)
        data = None
        try:
            async with async_timeout.timeout(5, loop=self.loop):
                called_url = URL(API_BASE_URL + endpoint_suffix, encoded=True)
                LOGGER.debug("Endpoint URL: %s", str(called_url))
                response = await self.session.get(url=called_url, headers=headers)
                if response.content_type.upper() == "APPLICATION/JSON":
                    data = await response.json()
                elif response.content_type.upper() == "APPLICATION/ZIP":
                    if response.headers.get('Transfer-Encoding') == 'chunked':
                        buffer = b""
                        async for dt, end_of_http_chunk in response.content.iter_chunks():
                            buffer += dt
                            if end_of_http_chunk:
                                data = buffer
                    else:
                        data = await response.read()
                else:
                    data = await response.read()
        except aiohttp.ClientError as error:
            LOGGER.error("Error connecting to STIB/MIVB API: %s", error)
        except asyncio.TimeoutError as error:
            LOGGER.debug("Timeout connecting to STIB/MIVB API: %s", error)
        return data

    async def close(self):
        """Close the session."""
        await self.session.close()



"""Setup tools config."""
import setuptools

with open("README.md", "r") as fh:
    LONG = fh.read()

setuptools.setup(
    name="pystibmivb",
    version="1.5.0",
    author="helldog136",
    author_email="dev.helldog136@outlook.com",
    python_requires=">=3.5.0",
    description=("Get realtime info on stop passages "
                 "of STIB/MIVB (opendata-api.stib-mivb.be)"),
    long_description=LONG,
    long_description_content_type="text/markdown",
    url="https://github.com/helldog136/pystibmivb",
    download_url="https://github.com/helldog136/pystibmivb/archive/1.5.0.tar.gz",
    packages=setuptools.find_packages(exclude=('tests',)),
    install_requires=[
        'async_timeout',
        'aiohttp',
        'pytz',
        'asyncio',
        'pyshp'
    ],
    license='MIT',
    classifiers=(
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
    ),
)



"""Initialize the package."""
from pystibmivb.client import AbstractSTIBAPIClient, STIBAPIClient
from pystibmivb.service import STIBService, InvalidLineFilterException, NoScheduleFromAPIException
from pystibmivb.service import ShapefileService
from .domain import *

NAME = "pystibmivb"